/*
   Hoarescope - Use Hoare logic for program correctness proofs
   Copyright (C) 2004 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/


#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <map>     // for pair<a,b>

#include "node.h"
#include "parser.h"

extern int yyerror(char *);

typedef struct proofTreeTag {
	nodeType *pre;
	nodeType *p;
	nodeType *post;
	char *rule;
	int id;
	int numproofs;
	struct proofTreeTag *proof[3];
} proofTree;

using namespace std;

typedef pair<nodeType *, string> alias_item;

vector<alias_item> aliases;

ostringstream texmsg;
extern ofstream texout;

vector<string> all_variables, assigned_variables, quantified_variables;

/* I have taken a lot of code from Thomas Niemann's excellent tutorial on
   lex and yacc, available from http://members.xoom.com/thomasn .

	Many thanks!

*/

int num_malloc = 0, num_free = 0;



nodeType *con(int i)
{
	nodeType *p = (nodeType *) malloc(sizeof(conNodeType));
	if (p == NULL) yyerror("out of memory");
	num_malloc++;
	p->type = typeCon;
	p->con.value = i;
	return p;
}

nodeType *id(char *s)
{
	nodeType *p = (nodeType *) malloc(sizeof(idNodeType));
	if (p == NULL) yyerror("out of memory");
	num_malloc++;
	p->type = typeId;
	p->id.str = (char *) malloc(strlen(s) + 1);
	if (p->id.str == NULL) yyerror("out of memory");
	strcpy(p->id.str, s);
	return p;
}

string node_to_string(nodeType *p);

nodeType *opr(int oper, int nops, ...)
{
	va_list ap;
	nodeType *p;
	int i;
	nodeType **ops;

	p = (nodeType *) malloc(sizeof(oprNodeType));
	if (p == NULL) yyerror("out of memory");
	num_malloc++;

	p->type = typeOpr;
	p->opr.oper = oper;
	p->opr.nops = nops;

	ops = (nodeType **) malloc(nops*sizeof(nodeType *));
	if (ops == NULL) yyerror("out of memory");

	va_start(ap, nops);
	for (i = 0; i < nops; i++) {
		ops[i] = va_arg(ap, nodeType *);
	}
	va_end(ap);
	p->opr.op = ops;
	return p;
}

proofTree *new_proof_tree(int id)
{
	proofTree *pt = (proofTree *) malloc(sizeof(proofTree));
	if (pt == NULL) yyerror("out of memory");

	num_malloc++;
	pt->pre = NULL; pt->p = NULL; pt->post = NULL;
	pt->rule = NULL;
	pt->numproofs = 0;
	pt->id = id;
	return pt;
}

void free_node(nodeType *p)
{
	if (!p) return;
	// cout << "Freeing: " << node_to_string(p) << endl;
	switch(p->type) {
	case typeId:
		if (p->id.str) free(p->id.str);
		break;
	case typeOpr:
		for (int i = 0; i < p->opr.nops; i++) {
			free_node(p->opr.op[i]);
		}
		free(p->opr.op);
		break;
	}
	free(p);
	num_free++;
}

void free_proof_tree(proofTree *pt)
{
	if (pt->pre != NULL) free_node(pt->pre);
	free_node(pt->p);
	if (pt->post != NULL) free_node(pt->post);
	for (int i = 0; i < pt->numproofs; i++) {
		free_proof_tree(pt->proof[i]);
	}
	free(pt);
	num_free++;
}


bool nodes_equal(nodeType *p1, nodeType *p2)
{
	if (p1 == NULL || p2 == NULL) return (p1 == p2);

	if (p1->type != p2 ->type) return false;

	switch(p1->type) {
	case typeId:
		return (strcmp(p1->id.str, p2->id.str) == 0);
	case typeCon:
		return p1->con.value == p2->con.value;
	case typeOpr:
		if (p1->opr.oper != p2->opr.oper) return false;
		if (p1->opr.nops != p2->opr.nops) return false;
		for (int i = 0; i < p1->opr.nops; i++) {
			if (!nodes_equal(p1->opr.op[i], p2->opr.op[i])) return false;
		}
		return true;
	}

	cout << "nodes_equal: cannot happen" << endl;
	return false;

}

nodeType *find_variable(char *v, nodeType *p)
{
	if (p == NULL) return NULL;
	switch (p->type) {
	case typeId:
		if (strcmp(p->id.str, v) == 0) return p;
		break;
	case typeOpr:
		for (int i = 0; i < p->opr.nops; i++) {
			if (find_variable(v, p->opr.op[i]) != NULL) return p->opr.op[i];
		}
	}
	return NULL;
}

void find_all_variables(nodeType *p)
{
	if (p == NULL) return;
	switch (p->type) {
	case typeId:
		all_variables.push_back(p->id.str);
		break;
	case typeOpr:
		for (int i = 0; i < p->opr.nops; i++) {
			find_all_variables(p->opr.op[i]);
		}
	}
}

void find_assigned_variables(nodeType *p)
{
	// saves a vector of all variables that are assigned values
	// in p. other variables can be regarded as constants
	if (p == NULL) return;
	if (p->type == typeOpr) {
		if (p->opr.oper == ASSIGN) {
			assigned_variables.push_back(p->opr.op[0]->id.str);
		} else {
			for (int i = 0; i < p->opr.nops; i++) {
				find_assigned_variables(p->opr.op[i]);
			}
		}
	}
}

void find_quantified_variables(nodeType *p)
{
	// saves a vector of all variables that are quantified in p.
	if (p == NULL) return;
	if (p->type == typeOpr) {
		if (p->opr.oper == EXISTS || p->opr.oper == FORALL) {
			quantified_variables.push_back(p->opr.op[0]->id.str);
		} else {
			for (int i = 0; i < p->opr.nops; i++) {
				find_quantified_variables(p->opr.op[i]);
			}
		}
	}
}


nodeType *copy_node(nodeType *p)
{
	// performs deep copy of node p
	nodeType **ops, *ret;

	if (p == NULL) return NULL;

	switch(p->type) {
	case typeCon:
		return con(p->con.value);
	case typeId:
		return id(p->id.str);
	case typeOpr:
		ops = (nodeType **) malloc(p->opr.nops * sizeof(nodeType *));
		if (ops == NULL) yyerror("out of memory");

		for (int i = 0; i < p->opr.nops; i++) {
			ops[i] = copy_node(p->opr.op[i]);
		}
		ret = (nodeType *) malloc(sizeof(oprNodeType));
		if (ret == NULL) yyerror("out of memory");
		num_malloc++;
		ret->opr.op = ops;
		ret->type = typeOpr;
		ret->opr.oper = p->opr.oper;
		ret->opr.nops = p->opr.nops;
		return ret;
	}
	return NULL; // cannot happen
}

void register_alias(nodeType *p, char *s)
{
	alias_item newitem(p, s);
	aliases.push_back(newitem);
}


string token_to_string(int token)
{
	switch (token) {
	case AND: return " AND ";
	case OR:  return " OR ";
	case EQ:  return " = ";
	case LE:  return " <= ";
	case GE:  return " >= ";
	case NE:  return " != ";
	case ASSIGN: return " := ";
	case IMPLIES: return " IMPLIES ";
	case WITH: return " ";
	case FORALL: return "FOR ALL ";
	case EXISTS: return "EXISTS ";
	default: return "COULD NOT CONVERT TOKEN";
	}

}

string node_to_string(nodeType *p)
{
	if (p == NULL) return "NULL";
	if (p->type == typeCon) {
		char temp_string[20];
		snprintf(temp_string, 20, "%d", p->con.value);
		return string(temp_string);
	}
	if (p->type == typeId) return p->id.str;

	switch (p->opr.oper) {

	case IF:
		return "if " + node_to_string(p->opr.op[0]) + " then " + node_to_string(p->opr.op[1]) + " else " + node_to_string(p->opr.op[2]);

	case EQ: case GE: case LE: case NE: case AND:
	case OR: case IMPLIES: case WITH:
		return "(" + node_to_string(p->opr.op[0]) + token_to_string(p->opr.oper) + node_to_string(p->opr.op[1]) + ")";

	case ASSIGN:
		return node_to_string(p->opr.op[0]) + token_to_string(p->opr.oper) + node_to_string(p->opr.op[1]);


	case WHILE:
		return "while " + node_to_string(p->opr.op[0]) + " do " + node_to_string(p->opr.op[1]);


	case COMPOUND:
		return "begin " + node_to_string(p->opr.op[0]) + "; " + node_to_string(p->opr.op[1]) + " end";

	case '+': case '-': case '<': case '>': case '*':
		return "(" + node_to_string(p->opr.op[0]) + " " + (char) p->opr.oper + " " + node_to_string(p->opr.op[1]) + ")";

	case NOT:
		return "NOT " + node_to_string(p->opr.op[0]);

	case FORALL: case EXISTS:
		return token_to_string(p->opr.oper) + node_to_string(p->opr.op[0]) + ": " + node_to_string(p->opr.op[1]);

	case SUBST:
		return "[" + node_to_string(p->opr.op[0]) + " <- " + node_to_string(p->opr.op[1]) + "]";
	}
	cout << "type was: " << p->opr.oper << endl;
	return "NOMATCH";
}


int next_free_id = 1;
int next_free_invariant = 1;

void prove(proofTree *);
nodeType *find_invariant(nodeType *, nodeType *, nodeType *);


void prove_compound(proofTree *pt)
{
	nodeType *op0 = pt->p->opr.op[0], *op1 = pt->p->opr.op[1];
	bool can_use_T1 = false;
	if (op0->opr.oper == ASSIGN) {
		char *variable = op0->opr.op[0]->id.str;
		// cout << "Can use T1? Variable: " << variable << endl;
		if (find_variable(variable, pt->pre) == NULL) {
			if (find_variable(variable, op0->opr.op[1]) == NULL) {
				can_use_T1 = true;
			} else  {
				// cout << "no: variable occurs in t" << endl;
			}
		} else {
			// cout << "no: variable occurs in PRE" << endl;
		}
	}

	if (can_use_T1) {
		int this_id = next_free_id++;
		cout << this_id << " (" << pt->id << ") T1: ";
		nodeType *intermed, *eq;
		eq = opr(EQ, 2, copy_node(op0->opr.op[0]), copy_node(op0->opr.op[1]));
		intermed = opr(AND, 2, copy_node(pt->pre), copy_node(eq));
		cout << node_to_string(intermed) << " {" << node_to_string(op1) << "} " << node_to_string(pt->post) << endl;

		proofTree *pt1 = new_proof_tree(this_id);
		pt1->pre = copy_node(intermed);
		pt1->p = copy_node(op1);
		pt1->post = copy_node(pt->post);
		pt->rule = "T1";
		pt->numproofs = 1;
		pt->proof[0] = pt1;

		prove(pt1);
		free_node(eq);
		free_node(intermed);
	} else if (op1->opr.oper == ASSIGN) {
		// T2
		int this_id = next_free_id++;
		cout << this_id << " (" << pt->id << ") T2: ";

		nodeType *t1 = opr(SUBST, 2, copy_node(op1->opr.op[0]), copy_node(op1->opr.op[1]));
		nodeType *t2 = opr(WITH, 2, copy_node(pt->post), copy_node(t1));
		cout << node_to_string(pt->pre) << " {" << node_to_string(op0) << "} " << node_to_string(t2) << endl;

		proofTree *pt2 = new_proof_tree(this_id);
		pt2->pre = copy_node(pt->pre);
		pt2->p = copy_node(op0);
		pt2->post = copy_node(t2);
		pt->numproofs = 1;
		pt->proof[0] = pt2;
		pt->rule = "T2";

		prove(pt2);
		free_node(t1); free_node(t2);
	} else if (op1->opr.oper == WHILE) {
		// T3
		nodeType *invariant = find_invariant(pt->pre, op1, pt->post);
		int id1 = next_free_id++;
		cout << id1 << " (" << pt->id << ") T3: ";
		cout << node_to_string(pt->pre) << " {" << node_to_string(op0) << "} " << node_to_string(invariant) << endl;

		proofTree *pt31, *pt32, *pt33;

		pt->numproofs = 3;
		pt->rule = "T3";

		pt31 = new_proof_tree(id1);
		pt31->pre = copy_node(pt->pre);
		pt31->p = copy_node(op0);
		pt31->post = copy_node(invariant);
		pt->proof[0] = pt31;

		prove(pt31);

		int id2 = next_free_id++;
		cout << id2 << " (" << pt->id << ") T3: ";
		nodeType *pre2 = opr(AND, 2, copy_node(invariant), copy_node(op1->opr.op[0]));
		cout << node_to_string(pre2) << " {" << node_to_string(op1->opr.op[1]) << "} " << node_to_string(invariant) << endl;

		pt32 = new_proof_tree(id2);
		pt32->pre = copy_node(pre2);
		pt32->p = copy_node(op1->opr.op[1]);
		pt32->post = copy_node(invariant);
		pt->proof[1] = pt32;

		prove(pt32);

		int id3 = next_free_id++;
		cout << id3 << " (" << pt->id << ") T3: ";
		nodeType *cond3 = opr(IMPLIES, 2, opr(AND, 2, copy_node(invariant), opr(NOT, 1, copy_node(op1->opr.op[0]))), copy_node(pt->post));
		cout << node_to_string(cond3) << endl;
		cout << "Left to prove: " << id3 << endl;

		pt33 = new_proof_tree(id3);
		pt33->p = copy_node(cond3);
		pt->proof[2] = pt33;

		free_node(invariant); free_node(pre2); free_node(cond3);
	} else {
		cout << "No rule for compound statement: " << pt->id << endl;
		if (op0->opr.oper == ASSIGN && !can_use_T1) {
			cout << "For your information: T1 is not allowed!" << endl;
		}
	}
}

void prove_if(proofTree *pt)
{
	int id1 = next_free_id++;
	cout << id1 << " (" << pt->id << ") ";
	nodeType *pre1, *pre2, *op0, *op1, *op2;
	op0 = pt->p->opr.op[0]; op1 = pt->p->opr.op[1]; op2 = pt->p->opr.op[2];

	pre1 = opr(AND, 2, copy_node(pt->pre), copy_node(op0));
	pre2 = opr(AND, 2, copy_node(pt->pre), opr(NOT, 1, copy_node(op0)));
	cout << "H3: " << node_to_string(pre1) << " {" << node_to_string(op1) << "} " << node_to_string(pt->post) << endl;

	int id2 = next_free_id++;
	cout << id2 << " (" << pt->id << ") H3: " << node_to_string(pre2) << " {" << node_to_string(op2) << "} " << node_to_string(pt->post) << endl;

	proofTree *pt1, *pt2;
	pt->rule = "H3";
	pt->numproofs = 2;

	pt1 = new_proof_tree(id1);
	pt1->pre = copy_node(pre1);
	pt1->p = copy_node(op1);
	pt1->post = copy_node(pt->post);
	pt->proof[0] = pt1;

	pt2 = new_proof_tree(id2);
	pt2->pre = copy_node(pre2);
	pt2->p = copy_node(op2);
	pt2->post = copy_node(pt->post);
	pt->proof[1] = pt2;

	prove(pt1);
	prove(pt2);
	free_node(pre1);
	free_node(pre2);
}


bool is_reflexive_assignment(nodeType *p, int tok)
{
	// reflexive assignment: x := x + c, x := x - c
	// these are useful for finding invariants
	if (p->type != typeOpr) return false;
	if (p->opr.oper != ASSIGN) return false;
	nodeType *var = p->opr.op[0];
	nodeType *value = p->opr.op[1];
	if (value->type != typeOpr) return false;
	if (value->opr.oper != tok) return false;
	nodeType *leftexpr = value->opr.op[0], *rightexpr = value->opr.op[1];
	if (leftexpr->type != typeId) return false;
	if (strcmp(var->id.str, leftexpr->id.str) != false) return false;

	if (rightexpr->type != typeCon) {
		if (rightexpr->type == typeId) {
			// in x := x +/- y, test if y is a quasi-constant
			// (variable that is not assigned a value)
			if (find(assigned_variables.begin(), assigned_variables.end(), rightexpr->id.str) != assigned_variables.end()) return false;
		} else {
			return false;
		}
	}

	return true;
}

string node_to_tex(nodeType *p, bool use_alias = true);

nodeType *find_invariant(nodeType *pre, nodeType *p, nodeType *post)
{

	char buf1[20], buf2[20];
	int invnum = next_free_invariant++;
	snprintf(buf1, 20, "I%d", invnum);
	snprintf(buf2, 20, "c%d", invnum);
	assigned_variables.clear();
	find_assigned_variables(p);

	if (nodes_equal(pre, p->opr.op[0])) {
		nodeType *i = opr(OR, 2, copy_node(pre), copy_node(post));
		cout << "Using invariant " << buf1 << ": " << node_to_string(i) << endl;

		texmsg << "Invariant $" << buf1 << "$: $" << node_to_tex(i) << "$\\hfil\\break" << endl;

		// return i;

		free_node(i);
		return id(buf1);
	}

	nodeType *exec = p->opr.op[1];
	if (exec->type == typeOpr && exec->opr.oper == COMPOUND) {
		// compound statement in while loop. check for
		// v := v + a; w := w - b    and
		// w := w - b; v := v + a

		nodeType *stmt1 = exec->opr.op[0], *stmt2 = exec->opr.op[1];
		/*
		cout << "stmt1: " << node_to_string(stmt1) << endl;
		if (is_reflexive_assignment(stmt1, '+')) {
			cout << "stmt1 is reflexive, PLUS" << endl;
		}
		if (is_reflexive_assignment(stmt1, '-'))
			cout << "stmt1 is reflexive, MINUS" << endl;

		if (is_reflexive_assignment(stmt2, '+'))
			cout << "stmt2 is reflexive, PLUS" << endl;

		if (is_reflexive_assignment(stmt2, '-'))
			cout << "stmt2 is reflexive, MINUS" << endl;
		*/

		nodeType *a, *b, *w, *v;
		if (is_reflexive_assignment(stmt1, '+') && is_reflexive_assignment(stmt2, '-')) {
			a = stmt1->opr.op[1]->opr.op[1];
			w = stmt2->opr.op[0];
			b = stmt2->opr.op[1]->opr.op[1];
			v = stmt1->opr.op[0];
			nodeType *invariant = opr(EQ, 2, opr('+', 2, opr('*', 2, copy_node(a), copy_node(w)), opr('*', 2, copy_node(b), copy_node(v))), id(buf2));
			cout << "Using invariant " << buf1 << ": " << node_to_string(invariant) << endl;
			nodeType *musthold = opr(IMPLIES, 2, copy_node(pre), copy_node(invariant));
			cout << "Left to do: Choose $" << buf2 << "$ so that $" << node_to_string(musthold) << "$ is valid in N" << endl;

			texmsg << "Invariant $" << buf1 << "$: $" << node_to_tex(invariant) << "$\\hfil\\break"<< endl;
			texmsg << "Choose $" << buf2 << "$ so that $" << node_to_tex(musthold) << "$ is valid in $\\mathbbm{N}$." << endl;
			
			free_node(musthold);
			free_node(invariant);
			return id(buf1);
		}

		if (is_reflexive_assignment(stmt1, '-') && is_reflexive_assignment(stmt2, '+')) {
			w = stmt1->opr.op[0];
			b = stmt1->opr.op[1]->opr.op[1];
			v = stmt2->opr.op[0];
			a = stmt2->opr.op[1]->opr.op[1];
			nodeType *invariant = opr(EQ, 2, opr('+', 2, opr('*', 2, copy_node(a), copy_node(w)), opr('*', 2, copy_node(b), copy_node(v))), id(buf2));
			cout << "Using invariant " << buf1 << ": " << node_to_string(invariant) << endl;
			nodeType *musthold = opr(IMPLIES, 2, copy_node(pre), copy_node(invariant));
			cout << "Left to do: Choose " << buf2 << " such that " << node_to_string(musthold) << " is valid in N" << endl;

			texmsg << "Invariant $" << buf1 << "$: $" << node_to_tex(invariant) << "$\\hfil\\break" << endl;
			texmsg << "Choose $" << buf2 << "$ so that $" << node_to_tex(musthold) << "$ is valid in $\\mathbbm{N}$." << endl;
			free_node(musthold);

			free_node(invariant);
			return id(buf1);
		}


		if ((is_reflexive_assignment(stmt1, '+') && is_reflexive_assignment(stmt2, '+')) || (is_reflexive_assignment(stmt1, '-') && is_reflexive_assignment(stmt2, '-'))) {
			v = stmt1->opr.op[0];
			w = stmt2->opr.op[0];
			a = stmt1->opr.op[1]->opr.op[1];
			b = stmt2->opr.op[1]->opr.op[1];

			nodeType *invariant1 = opr(EQ, 2, opr('*', 2, copy_node(a), copy_node(w)), opr('+', 2, opr('*', 2, copy_node(b), copy_node(v)), id(buf2)));

			nodeType *invariant2 = opr(EQ, 2, opr('+', 2, opr('*', 2, copy_node(a), copy_node(w)), id(buf2)), opr('*', 2, copy_node(b), copy_node(v)));

			nodeType *condition1 = opr(IMPLIES, 2, copy_node(pre), copy_node(invariant1));
			nodeType *condition2 = opr(IMPLIES, 2, copy_node(pre), copy_node(invariant2));

			cout << "I can not decide which invariant to use. Choices are: " << endl;
			cout << "[1] " << node_to_string(invariant1) << endl;
			cout << "[2] " << node_to_string(invariant2) << endl;
			cout << "Left to do: Choose " << buf2 << " so that" << endl;
			cout << node_to_string(condition1) << "      or" << endl;
			cout << node_to_string(condition2) << "     (respectively) " << endl << " are valid in N and ";
			cout << buf2 << " is positive. For this invariant I will write " << buf1 << endl;


			texmsg << "Invariant $" << buf1 << "$ is either: $" << node_to_tex(invariant1) << "$ or: $" << node_to_tex(invariant2) << "$; $" << buf2 << "$ should be positive, and $" << node_to_tex(condition1) << "$ or $" << node_to_tex(condition2) << "$ should hold, respectively." << endl;

			free_node(invariant1); free_node(invariant2);
			free_node(condition1); free_node(condition2);
			return id(buf1);
		}
	}

	cout << "Could not find invariant - using symbolic name " << buf1 << endl;
	texmsg << "Invariant $" << buf1 << "$ could not be computed." << endl;
	return id(buf1);
}

void prove_while(proofTree *pt)
{
	//cout << id1 << " (" << parent << ") ";
	nodeType *t, *invariant = find_invariant(pt->pre, pt->p, pt->post);
	pt->numproofs = 3;
	pt->rule = "H4";


	int id1 = next_free_id++;
	t = opr(IMPLIES, 2, copy_node(pt->pre), copy_node(invariant));
	cout << id1 << " (" << pt->id << ") ";
	cout << "H4: " << node_to_string(t) << endl;
	cout << "Left to prove: " << id1 << endl;

	proofTree *pt1 = new_proof_tree(id1);
	pt1->p = copy_node(t);
	pt->proof[0] = pt1;

	free_node(t);

	int id2 = next_free_id++;
	nodeType *pre1;
	pre1 = opr(AND, 2, copy_node(invariant), copy_node(pt->p->opr.op[0]));
	cout << id2 << " (" << pt->id << ") ";
	cout << "H4: " << node_to_string(pre1) << " {" << node_to_string(pt->p->opr.op[1]) << "} " << node_to_string(invariant) << endl;

	proofTree *pt2 = new_proof_tree(id2);
	pt2->pre = copy_node(pre1);
	pt2->p = copy_node(pt->p->opr.op[1]);
	pt2->post = copy_node(invariant);
	pt->proof[1] = pt2;


	int id3 = next_free_id++;
	t = opr(IMPLIES, 2, opr(AND, 2, copy_node(invariant), opr(NOT, 1, copy_node(pt->p->opr.op[0]))), copy_node(pt->post));
	cout << id3 << " (" << pt->id << ") ";
	cout << "H4: " << node_to_string(t) << endl;
	cout << "Left to prove: " << id3 << endl;

	proofTree *pt3 = new_proof_tree(id3);
	pt3->p = copy_node(t);
	pt->proof[2] = pt3;

	free_node(t);


	prove(pt2);
	free_node(invariant);
	free_node(pre1);
}

void prove_assign(proofTree *pt) {
	int this_id = next_free_id++;

	quantified_variables.clear();
	find_quantified_variables(pt->pre);
	find_quantified_variables(pt->post);

	all_variables.clear();
	find_all_variables(pt->p->opr.op[0]);
	find_all_variables(pt->p->opr.op[1]);

	for (int i = 0; i < all_variables.size(); i++) {
		if (find(quantified_variables.begin(), quantified_variables.end(), all_variables[i]) != quantified_variables.end()) {
			cout << "Variable " << all_variables[i] << " is quantified, H1 not possible" << endl;
			return;
		}
	}

	nodeType *n1, *n2;
	cout << this_id << " (" << pt->id << ") H1: ";
	n1 = opr(SUBST, 2, copy_node(pt->p->opr.op[0]), copy_node(pt->p->opr.op[1]));
	n2 = opr(IMPLIES, 2, copy_node(pt->pre), opr(WITH, 2, copy_node(pt->post), copy_node(n1)));
	cout << node_to_string(n2) << endl;
	cout << "Left to prove: " << this_id << endl;

	proofTree *final = new_proof_tree(this_id);
	final->p = copy_node(n2);

	pt->rule = "H1";
	pt->numproofs = 1;
	pt->proof[0] = final;

	free_node(n1); free_node(n2);
}


void prove(proofTree *pt)
{
	// cout << "# Inference of " << node_to_string(pre) << " {" << node_to_string(p) << "} " << node_to_string(post) << ": " << endl;
	switch (pt->p->opr.oper) {
	case ASSIGN: prove_assign(pt);
		break;

	case COMPOUND: prove_compound(pt);
		break;

	case IF: prove_if(pt);
		break;

	case WHILE: prove_while(pt);
		break;

	default:
		cout << "No rule for: " << node_to_string(pt->p) << endl;
	}
}


string token_to_tex(int token)
{
	switch (token) {
	case AND: return " \\land ";
	case OR:  return " \\lor ";
	case EQ:  return " = ";
	case LE:  return " \\le ";
	case GE:  return " \\ge ";
	case NE:  return " \\neq ";
	case '-': return " \\dotminus ";
	case ASSIGN: return " \\gets ";
	case IMPLIES: return " \\supset ";
	case WITH: return " ";
	case FORALL: return " \\forall ";
	case EXISTS: return " \\exists ";
	default: return "COULD NOT CONVERT TOKEN";
	}

}

string rm(string s)
{
	return "{\\rm " + s + "}";
}

string node_to_tex(nodeType *p, bool use_alias)
{
	if (p == NULL) return "NULL";

	if (use_alias) {
		// for defined alias, return this alias
		for (int i = 0; i < aliases.size(); i++) {
			if (nodes_equal(aliases[i].first, p)) return aliases[i].second;
		}
	}

	if (p->type == typeCon) {
		char temp_string[20];
		snprintf(temp_string, 20, "%d", p->con.value);
		return string(temp_string);
	}
	if (p->type == typeId) return rm(p->id.str);

	switch (p->opr.oper) {

	case IF:
		return "\\stmtif " + node_to_tex(p->opr.op[0]) + "\\stmtthen " + node_to_tex(p->opr.op[1]) + " \\stmtelse " + node_to_tex(p->opr.op[2]);

	case EQ: case GE: case LE: case NE: case AND:
	case OR: case IMPLIES: case WITH: case '-':
		return "(" + node_to_tex(p->opr.op[0]) + token_to_tex(p->opr.oper) + node_to_tex(p->opr.op[1]) + ")";

	case ASSIGN:
		return node_to_tex(p->opr.op[0]) + token_to_tex(p->opr.oper) + node_to_tex(p->opr.op[1]);


	case WHILE:
		return "\\stmtwhile " + node_to_tex(p->opr.op[0]) + "\\stmtdo " + node_to_tex(p->opr.op[1]);


	case COMPOUND:
		return "\\stmtbegin " + node_to_tex(p->opr.op[0]) + rm(";\\; ") + node_to_tex(p->opr.op[1]) + "\\stmtend ";

	case '+': case '<': case '>': case '*':
		return "(" + node_to_tex(p->opr.op[0]) + " " + (char) p->opr.oper + " " + node_to_tex(p->opr.op[1]) + ")";

	case NOT:
		return "\\lnot " + node_to_tex(p->opr.op[0]);

	case FORALL: case EXISTS:
		return "(" + token_to_tex(p->opr.oper) + node_to_tex(p->opr.op[0]) + ") " + node_to_tex(p->opr.op[1]);

	case SUBST:
		return "\\subst{" + node_to_tex(p->opr.op[0]) + "}{" + node_to_tex(p->opr.op[1]) + "}";
	}
	cout << "type was: " << p->opr.oper << endl;
	return "NOMATCH";
}

int next_inference_number = 1;

string tree_to_tex(proofTree *pt)
{
	string thisline;
	if (pt->pre != NULL) {
		thisline = node_to_tex(pt->pre) + " \\;\\{" + node_to_tex(pt->p) + "\\}\\; " + node_to_tex(pt->post);
	} else {
		thisline = node_to_tex(pt->p);
	}
	if (pt->numproofs == 0) {
		char buf[20];
		snprintf(buf, 20, "{(%d)}", next_inference_number++);
		return "\n\\deduce{" + thisline + "}" + buf;
	}


	// the explicit conversion of pt->rule to "string" is
	// necessary for my compiler
	string ret = "\n\\infer[(" + rm(string(pt->rule)) + ")]{" + thisline + "}{";

	for (int i = 0; i < pt->numproofs; i++) {
		ret = ret + tree_to_tex(pt->proof[i]);
		if (i < pt->numproofs - 1) ret = ret + "\n&";
	}

	return ret + "}";
}


void do_tex_output(proofTree *pt)
{
	next_inference_number = 1;

	if (!aliases.empty()) {
		texout << "\\noindent The following aliases were used:" << endl;
		for (int i = 0; i < aliases.size(); i++) {
			alias_item aitem = aliases[i];
			texout << "     \\hfil\\break" << endl << "$" << aitem.second << "\\equiv " << node_to_tex(aitem.first, false) << "$" << endl;
		}
		texout << "\\hfil\\vspace{1cm}" << endl;
	}
	texout << tree_to_tex(pt) << endl;
	texout << endl << "\\vspace{1cm}" << endl << endl;
	texout << texmsg.str() << endl;
	texout << endl << "\\vspace{2cm}" << endl << endl;
}


void prove_program(nodeType *pre, nodeType *p, nodeType *post)
{
	next_free_id = 1;
	next_free_invariant = 1;

	texmsg.str("");

	proofTree *ptree = new_proof_tree(0);
	ptree->pre = pre; ptree->p = p; ptree->post = post;
	/* cout << "I have to prove: " << node_to_string(p) << endl;
	cout << "pre: " << node_to_string(pre) << " ; post: " << node_to_string(post) << endl; */
	prove(ptree);
	if (texout) do_tex_output(ptree);
	cout << "End of inference" << endl << endl;
	free_proof_tree(ptree);

	// clear alias list
	for (int i = 0; i < aliases.size(); i++) {
		free_node(aliases[i].first);
	}
	aliases.clear();

	if (num_malloc != num_free) {
		cout << "# Internal problem: num_malloc: " << num_malloc << ", num_free: " << num_free << endl;
	}
}



/* What follows is code that one could use for an interpreter of AL;
   we don't need it right now, but it could be used for some kinds
   of tests.
*/


/*

int ex(nodeType *);
extern map<string, string> variables;

void doVariableAssignment(nodeType *p)
{
	string variable = p->opr.op[0]->id.str;

	switch(p->opr.op[1]->type) {
	case typeId:
		variables[variable] = variables[p->opr.op[1]->id.str];
		break;
	case typeCon:
	case typeOpr:
		int value = ex(p->opr.op[1]);
		char buf[20];
		snprintf(buf, 20, "%d", value);
		variables[variable] = buf;
		break;
	}
}


int ex(nodeType *p)
{
	if (p->type == typeCon) return p->con.value;
	if (p->type == typeId)  return atoi(variables[p->id.str].c_str());

	switch (p->opr.oper) {

	case IF:
		if (ex(p->opr.op[0])) {
			return (ex(p->opr.op[1]));
		} else {
			return (ex(p->opr.op[2]));
		}
		break;

	case EQ: return (ex(p->opr.op[0]) == ex(p->opr.op[1]));

	case GE: return (ex(p->opr.op[0]) >= ex(p->opr.op[1]));

	case LE: return (ex(p->opr.op[0]) <= ex(p->opr.op[1]));

	case NE: return (ex(p->opr.op[0]) != ex(p->opr.op[1]));

	case WHILE:
		while (ex(p->opr.op[0])) {
			if (ex(p->opr.op[1]) == -1) return -1;
		}
		return 0;

	case '+': return (ex(p->opr.op[0]) + ex(p->opr.op[1]));
	case '-': return (ex(p->opr.op[0]) - ex(p->opr.op[1]));
	case '<': return (ex(p->opr.op[0]) < ex(p->opr.op[1]));
	case '>': return (ex(p->opr.op[0]) > ex(p->opr.op[1]));
	case ASSIGN: doVariableAssignment(p); break;
	case COMPOUND: if (ex(p->opr.op[0]) == -1) {
			return -1;
		} else {
			return ex(p->opr.op[1]);
		}
	}
	return 0;
}

*/
